import numpy as np
from dataclasses import dataclass
from enum import Enum


class ConstType(Enum):
    DROP = 1
    QR = 2
    DIFF = 3

@dataclass
class Constraint:
  """
  Constraint storage. ``Z``, either holds the Qr-based correction matrix that needs
  to be multiplied with :math:`\mathbf{X}`, :math:`\mathbf{S}`, and :math:`\mathbf{D}` (where :math:`\mathbf{D}\mathbf{D}^T = \mathbf{S}`) to make terms subject to the conventional sum-to-zero constraints
  applied also in mgcv (Wood, 2017), the column/row that should be dropped from those - then :math:`\mathbf{X}` can also no longer take
  on a constant, or ``None`` indicating that the model should be "difference re-coded" to enable sparse sum-to-zero
  constraints. The latter two are available in mgcv's ``smoothCon`` function by setting the ``sparse.cons`` argument to 1 or
  2 respectively.

  The QR-based approach is described in detail by Wood (2017) and is similar to just mean centering every basis function involved in the
  smooth and then dropping one column from the corresponding centered model matrix. The column-dropping approach is self-explanatory. The
  difference re-coding deserves some explanation. Consider the model:

  .. math::

    y = f(x) = b_1*f_1(x) + b_2*f_2(x) + b_3*f_3(x) + b_4*f_4(x) + b_5*f_5(x)
  
  In ``mssm`` :math:`f(x)` will be parameterized via five B-spline basis functions :math:`f_i(x)`, each weighted by a coefficient :math:`b_i`. Each of the :math:`f_i(x)` has
  support only over a small range of :math:`x` - resulting in a sparse model matrix :math:`\mathbf{X}`. This desirable property is lost when enforcing the conventional
  sum-to-zero constraints (either QR-based or by subtracting the mean of each basis over :math:`x` from the corresponding :math:`f_i(x)`). To maintain sparsity,
  one can enforce a coefficients-sum-to-zero constraint (Wood, 2020) by re-coding instead:

  Set :math:`b_1 = -b_2`

  Then, re-write:
  
  .. math::

    y = -b_2 * f_1(x) + b_2 * f_2(x) + b_3 * f_3(x) - b_3 * f_2(x) + b_4 * f_4(x) - b_4 * f_3(x) + b_5 * f_5(x) - b_5 * f_4(x)

    y = -b_2 * f_1(x) + (b_2 - b_3) * f_2(x) + (b_3 - b_4) * f_3(x) + (b_4 - b_5) * f_4(x) + b_5 * f_5(x)

    y = b_2 * (f_2(x) - f_1(x)) + b_3 * (f_3(x) - f_2(x)) + b_4 * (f_4(x) - f_3(x)) + b_5 * (f_5(x) - f_4(x))

  Line 3 shows how the constraint can be absorbed for model fitting by first computing new bases :math:`(f_i(x) - f_{i-1}(x))` and then estimating
  the coefficients based on those (this is implemented in mgcv's ``smoothCon`` function when setting ``sparse.cons=2``). Note that the constrained
  model, just like when using the QR-based or dropping approach, requires dropping one of the :math:`k` original coefficients for estimation since only :math:`k-1`
  coefficients are allowed to vary. The same sum-to-zero constraint can be achieved by fixing one of the central bases in the original model to it's neighbor (e.g., setting :math:`b_2 = -b_3`) or
  by setting :math:`b_1= -b_2 - b_3 - b_4 - b_5`. ``mssm`` fixes one of the central bases to it's neighbor.

  With a B-splines basis, it would be necessary that :math:`b_1 = b_2 = b_3 = b_4 = b_5` for the model to fit a constant :math:`f(x)` over all values of :math:`x` (Eilers & Marx, 1996).
  In the constrained model this is no longer possible due to :math:`b_1 = -b_2`, effectively removing the constant from the space of functions that :math:`f(x)` can approximate.
  Instead, in the constrained model if :math:`b_2 = b_3 = b_4 = b_5`, :math:`f(x)` summed over all :math:`x` will be equal to zero!
  
  Importantly, the "new" bases obtained by computing :math:`(f_i(x) - f_{i-1}(x))` are all still relatively sparse. This is because neighboring
  :math:`f_i(x)` overlap in their support and because the support of each individual :math:`f_i(x)` is narrow, i.e., :math:`f_i(x)` is sparse itself.

  However, this is not a true centering constraint: :math:`f(x)` will not necessarily be orthogonal to the intercept, i.e., :math:`\mathbf{1}^T \mathbf{f(x)}` will not necessarily be 0. Hence, confidence intervals will usually
  be wider when using ConstType.DIFF (also when using ConstType.DROP, for the same reason) instead of ``ConstType.QR`` (see Wood; 2017,2020)!

  A final note regards the use of tensor smooths when ``te==False``. Since the value of any constant estimated for a smooth depends on the type of constraint used, the mmarginal functions estimated
  for the "main effects" (:math:`f(x)`, :math:`f(z)`) and "interaction effect" (:math:`f(x,z)`) in a model: :math:`y = a + f(x) + f(z) + f(x,z)` will differ depending on the type of constraint used. The "Anova-like" decomposition
  described in detail in Wood (2017) is achievable only when using ``ConstType.QR``.

  References:
    - Wood, S. N. (2017). Generalized Additive Models: An Introduction with R, Second Edition (2nd ed.).
    - Wood, S. N. (2020). Inference and computation with generalized additive models and their extensions. TEST, 29(2), 307–339. https://doi.org/10.1007/s11749-020-00711-5
    - Eilers, P. H. C., & Marx, B. D. (1996). Flexible smoothing with B-splines and penalties. Statistical Science, 11(2), 89–121. https://doi.org/10.1214/ss/1038425655

  """
  Z: np.array or int = None
  type:ConstType = None


from mssm.models import *
import numpy as np
import os
from mssmViz.sim import*
import io
from contextlib import redirect_stdout


class Test_GAUMLS:
    # Simulate 500 data points
    GAUMLSSDat = sim6(500,seed=20)

    # We need to model the mean: \mu_i = \alpha + f(x0)
    formula_m = Formula(lhs("y"),
                        [i(),f(["x0"],nk=10)],
                        data=GAUMLSSDat)

    # and the standard deviation as well: log(\sigma_i) = \alpha + f(x0)
    formula_sd = Formula(lhs("y"),
                        [i(),f(["x0"],nk=10)],
                        data=GAUMLSSDat)

    # Collect both formulas
    formulas = [formula_m,formula_sd]

    # Create Gaussian GAMMLSS family with identity link for mean
    # and log link for sigma
    family = GAUMLSS([Identity(),LOG()])

    # Now define the model and fit!
    model = GAMMLSS(formulas,family)
    model.fit(extension_method_lam="nesterov",max_inner=50,min_inner=50)

    def test_GAMedf(self):
        assert round(self.model.edf,ndigits=3) == 18.358

    def test_GAMcoef(self):
        coef = self.model.overall_coef.flatten()
        assert np.allclose(coef,np.array([ 3.58432714, -9.20689268, -0.3901218 ,  4.33485776, -2.83325345,
                                            -4.68428463, -1.93389004, -4.14504997, -6.70579839, -4.7240103 ,
                                            -5.19939664,  0.02026291, -1.35759472,  1.5379936 ,  2.39353569,
                                            2.33544107,  2.1749611 ,  2.3508268 ,  1.97975281,  1.4404345 ,
                                            -0.96502575, -4.28148962])) 

    def test_GAMlam(self):
        lam = np.array([p.lam for p in self.model.overall_penalties])
        assert np.allclose(lam,np.array([0.00909086, 1.07657659])) 

    def test_GAMreml(self):
        reml = self.model.get_reml()
        assert round(reml,ndigits=3) == -772.602 

    def test_GAMllk(self):
        llk = self.model.get_llk(False)
        assert round(llk,ndigits=3) == -719.601 
    
    def test_print_smooth(self):
        capture = io.StringIO()
        with redirect_stdout(capture):
            self.model.print_smooth_terms()
        capture = capture.getvalue()

        comp = "\nDistribution parameter: 1\n\nf(['x0']); edf: 9.799\n\nDistribution parameter: 2\n\nf(['x0']); edf: 6.559\n"
        assert comp == capture

    def test_print_parametric(self):
        capture = io.StringIO()
        with redirect_stdout(capture):
            self.model.print_parametric_terms()
        capture = capture.getvalue()

        comp = "\nDistribution parameter: 1\n\nIntercept: 3.584, z: 58.435, P(|Z| > |z|): 0.000e+00 ***\n\nNote: p < 0.001: ***, p < 0.01: **, p < 0.05: *, p < 0.1: .\n\nDistribution parameter: 2\n\nIntercept: 0.02, z: 0.64, P(|Z| > |z|): 0.52195\n\nNote: p < 0.001: ***, p < 0.01: **, p < 0.05: *, p < 0.1: .\n"
        assert comp == capture
   
    def test_print_smooth_p(self):
        capture = io.StringIO()
        with redirect_stdout(capture):
            self.model.print_smooth_terms(p_values=True)
        capture = capture.getvalue()

        comp = "\nDistribution parameter: 1\n\nf(['x0']); edf: 9.799 chi^2: 5768.018 P(Chi^2 > chi^2) = 0.000e+00 ***\n\nNote: p < 0.001: ***, p < 0.01: **, p < 0.05: *, p < 0.1: . p-values are approximate!\n\nDistribution parameter: 2\n\nf(['x0']); edf: 6.559 chi^2: 563.181 P(Chi^2 > chi^2) = 0.000e+00 ***\n\nNote: p < 0.001: ***, p < 0.01: **, p < 0.05: *, p < 0.1: . p-values are approximate!\n"
        assert comp == capture

class Test_GAUMLS_MIXED:
    ## Simulate some data - effect of x0 on scale parameter is very very small
    sim_fit_dat = sim10(n=500,c=0.1,seed=20)

    # Now fit nested models

    # Create Gaussian GAMMLSS family with identity link for mean
    # and log link for sigma
    family = GAUMLSS([Identity(),LOGb(-0.3)])

    # We need to model the mean: \mu_i = \alpha + f(x0)
    sim1_formula_m = Formula(lhs("y"),
                        [i(),f(["x0"]),f(["x1"]),ri("x4")],
                        data=sim_fit_dat)

    # and the standard deviation as well: log(\sigma_i) = \alpha + f(x0)
    sim1_formula_sd = Formula(lhs("y"),
                        [i(),f(["x2"]),f(["x3"])],
                        data=sim_fit_dat)

    # Collect both formulas
    sim1_formulas = [sim1_formula_m,sim1_formula_sd]

    model = GAMMLSS(sim1_formulas,family)
    model.fit(seed=30,max_outer=250,max_inner=500,extend_lambda=True,method="QR/Chol")

    def test_GAMedf(self):
        assert round(self.model.edf,ndigits=3) == 26.196

    def test_GAMlam(self):
        lam = np.array([p.lam for p in self.model.overall_penalties])
        assert np.allclose(lam,np.array([1.00000000e+07, 6.04896788e+00, 5.14594390e-01, 9.12400516e-01,5.38283393e+02])) 

    def test_GAMreml(self):
        reml = self.model.get_reml()
        assert round(reml,ndigits=3) == -1747.292

    def test_GAMllk(self):
        llk = self.model.get_llk(False)
        assert round(llk,ndigits=3) == -1715.506

class Test_GAMMALS:

    # Simulate 500 data points
    GAMMALSSDat = sim6(500,family=GAMMALS([LOG(),LOG()]),seed=10)

    # We need to model the mean: \mu_i = \alpha + f(x0)
    formula_m = Formula(lhs("y"),
                        [i(),f(["x0"],nk=10)],
                        data=GAMMALSSDat)

    # and the scale parameter as well: log(\scale_i) = \alpha + f(x0)
    formula_sd = Formula(lhs("y"),
                        [i(),f(["x0"],nk=10)],
                        data=GAMMALSSDat)

    # Collect both formulas
    formulas = [formula_m,formula_sd]

    # Create Gamma GAMMLSS family with log link for mean
    # and log link for sigma
    family = GAMMALS([LOG(),LOG()])

    # Now define the model and fit!
    model = GAMMLSS(formulas,family)
    model.fit(extension_method_lam="nesterov",max_inner=50,min_inner=50)

    def test_GAMedf(self):
        assert round(self.model.edf,ndigits=3) == 14.677 

    def test_GAMcoef(self):
        coef = self.model.overall_coef.flatten()
        assert np.allclose(coef,np.array([ 1.14440894, -0.79166857,  1.95574948,  2.46973458,  1.7725019 ,
                                            1.29047048,  1.89204546,  1.34016392,  0.24865949, -0.66762416,
                                            -1.696964  ,  0.7715949 , -0.72928057,  0.87660987,  1.12963098,
                                            1.22774166,  1.1121008 ,  1.27841939,  1.38510189,  0.72797382,
                                            0.02458096, -0.55655912])) 

    def test_GAMlam(self):
        lam = np.array([p.lam for p in self.model.overall_penalties])
        assert np.allclose(lam,np.array([0.51599945, 2.4946028])) 

    def test_GAMreml(self):
        reml = self.model.get_reml()
        assert round(reml,ndigits=3) == -923.897 

    def test_GAMllk(self):
        llk = self.model.get_llk(False)
        assert round(llk,ndigits=3) == -892.042

class Test_mulnom_hard:
    # Test multinomial model

    # We need to specify K-1 formulas - see the `MULNOMLSS` docstring for details.
    formulas = [Formula(lhs("y"),
                        [i(),f(["x0"])],
                        data=sim5(1000,seed=91)) for k in range(4)]

    # Create family - again specifying K-1 pars - here 4!
    family = MULNOMLSS(4)

    # Fit the model
    model = GAMMLSS(formulas,family)
    model.fit(method="Chol",control_lambda=0,extend_lambda=False,max_outer=200,max_inner=500,should_keep_drop=False)

    def test_GAMedf(self):
        assert round(self.model.edf,ndigits=3) == 15.114

    def test_GAMcoef(self):
        coef = self.model.overall_coef.flatten()
        assert np.allclose(coef,np.array([ 1.32831232, -0.70411299,  0.26038397,  0.8585649 ,  1.31636532,
                                            1.46064575,  1.15467552,  0.34090563, -0.96911006, -2.2518714 ,
                                            0.66532703, -0.76857181, -0.20872371,  0.11240685,  0.42769554,
                                            0.70741374,  1.07469267,  1.32593173,  1.37436662,  1.48506959,
                                            2.06156634, -1.63789937,  7.86561269,  9.15961217,  4.10367107,
                                            1.21906228,  1.68896079,  1.01054086, -0.31833491, -0.59758569,
                                            0.4881652 , -0.46708693, -0.12681357,  0.06836815,  0.25998144,
                                            0.42995532,  0.65313117,  0.80576918,  0.83513642,  0.90234416])) 

    def test_GAMlam(self):
        lam = np.array([p.lam for p in self.model.overall_penalties])
        assert np.allclose(lam,np.array([3.37059261e+00, 7.24169569e+04, 5.02852103e-02, 1.00000000e+07]))

    def test_GAMreml(self):
        reml = self.model.get_reml()
        assert round(reml,ndigits=3) == -1002.689
    
    def test_GAMllk(self):
        llk = self.model.get_llk(False)
        assert round(llk,ndigits=3) == -980.676


class Test_mulnom_hard_repara:
    # Test multinomial model

    # We need to specify K-1 formulas - see the `MULNOMLSS` docstring for details.
    formulas = [Formula(lhs("y"),
                        [i(),f(["x0"])],
                        data=sim5(1000,seed=91)) for k in range(4)]

    # Create family - again specifying K-1 pars - here 4!
    family = MULNOMLSS(4)

    # Fit the model
    model = GAMMLSS(formulas,family)
    model.fit(method="Chol",control_lambda=0,extend_lambda=False,max_outer=200,max_inner=500,should_keep_drop=False,repara=True)

    def test_GAMedf(self):
        assert round(self.model.edf,ndigits=3) == 15.114

    def test_GAMcoef(self):
        coef = self.model.overall_coef.flatten()
        assert np.allclose(coef,np.array([ 1.32831232, -0.70411299,  0.26038397,  0.85856489,  1.31636532,
                                            1.46064575,  1.15467552,  0.34090563, -0.96911006, -2.2518714 ,
                                            0.66532703, -0.76857181, -0.20872371,  0.11240685,  0.42769554,
                                            0.70741373,  1.07469267,  1.32593173,  1.37436662,  1.48506958,
                                            2.06156634, -1.63789936,  7.86561269,  9.15961218,  4.10367108,
                                            1.21906229,  1.6889608 ,  1.01054087, -0.31833491, -0.5975857 ,
                                            0.4881652 , -0.46708693, -0.12681357,  0.06836815,  0.25998144,
                                            0.42995532,  0.65313117,  0.80576918,  0.83513642,  0.90234416])) 

    def test_GAMlam(self):
        lam = np.array([p.lam for p in self.model.overall_penalties])
        assert np.allclose(lam,np.array([3.37059261e+00, 7.24170552e+04, 5.02852109e-02, 1.00000000e+07]))

    def test_GAMreml(self):
        reml = self.model.get_reml()
        assert round(reml,ndigits=3) == -1000.749
    
    def test_GAMllk(self):
        llk = self.model.get_llk(False)
        assert round(llk,ndigits=3) == -980.676
     

class Test_pred_whole_func_cor:
    # Simulate some data
    sim_fit_dat = sim6(n=500,seed=37)

    # Now fit nested models

    # Create Gaussian GAMMLSS family with identity link for mean
    # and log link for sigma
    family = GAUMLSS([Identity(),LOG()])

    # We need to model the mean: \mu_i = \alpha + f(x0)
    sim1_formula_m = Formula(lhs("y"),
                        [i(),f(["x0"],nk=15)],
                        data=sim_fit_dat)

    # and the standard deviation as well: log(\sigma_i) = \alpha + f(x0)
    sim1_formula_sd = Formula(lhs("y"),
                        [i(),f(["x0"],nk=15)],
                        data=sim_fit_dat)

    # Collect both formulas
    sim1_formulas = [sim1_formula_m,sim1_formula_sd]

    sim_fit_model = GAMMLSS(sim1_formulas,family)
    sim_fit_model.fit(extension_method_lam="nesterov")

    # Test prediction code + whole-function correction
    pred_dat = pd.DataFrame({"x0":np.linspace(0,1,50)})
    pred,pred_mat,ci = sim_fit_model.predict(0,[0,1],pred_dat,ci=True,whole_interval=True,seed=20)

    def test_pred(self):
        assert np.allclose(self.pred,np.array([-0.01138392,  0.05310652,  0.44636549,  1.15072365,  2.14835029,
                                                3.39365478,  4.75023087,  6.06313038,  7.17850177,  8.00856086,
                                                8.56808424,  8.88068926,  8.96949259,  8.8500654 ,  8.53214278,
                                                8.02530808,  7.34686356,  6.5607974 ,  5.74856022,  4.99160243,
                                                4.35102523,  3.83236853,  3.43177651,  3.14527921,  2.9643237 ,
                                                2.87431683,  2.86025298,  2.90661517,  2.99179483,  3.09018223,
                                                3.17609768,  3.22449593,  3.21350857,  3.12225943,  2.92994466,
                                                2.63296809,  2.26712914,  1.87371669,  1.49344223,  1.15055278,
                                                0.85083955,  0.5991164 ,  0.39941563,  0.24828419,  0.13810493,
                                                0.06121592,  0.01075972, -0.01677633, -0.0240358 , -0.01348688]))
    
    def test_ci(self):
        assert np.allclose(self.ci,np.array([0.27251711, 0.11885948, 0.14575675, 0.20026483, 0.26016534,
                                                0.27019453, 0.31642028, 0.39883243, 0.43331877, 0.44002722,
                                                0.50908004, 0.59067928, 0.5932289 , 0.57954789, 0.64498692,
                                                0.71516238, 0.69827735, 0.69170816, 0.76685076, 0.82708777,
                                                0.82194123, 0.87061588, 0.97880911, 1.01529647, 0.96096182,
                                                0.93911891, 0.96154628, 0.91487953, 0.8149885 , 0.7760991 ,
                                                0.77936308, 0.71521874, 0.6455595 , 0.65966908, 0.67793687,
                                                0.61339351, 0.55758266, 0.57039461, 0.56076992, 0.49618751,
                                                0.47606143, 0.50317253, 0.46317742, 0.36127176, 0.30228488,
                                                0.28955845, 0.22256101, 0.15345764, 0.10679976, 0.2008102 ]))
        

class Test_diff_whole_func_cor:
    # Difference prediction for gammlss

    # Simulate some data - effect of x0 on scale parameter is very very small in one condition
    sim_fit_dat1 = sim8(n=500,c=0.1,seed=37)
    sim_fit_dat1["cond"] = "a"
    sim_fit_dat2 = sim8(n=500,c=1,seed=41)
    sim_fit_dat2["cond"] = "b"
    sim_fit_dat = pd.concat((sim_fit_dat1,sim_fit_dat2),axis=0,ignore_index=True)

    # Now fit model

    # Create Gaussian GAMMLSS family with identity link for mean
    # and log link for sigma
    family = GAUMLSS([Identity(),LOG()])

    # We need to model the mean: \mu_i = \alpha + f(x0)
    sim1_formula_m = Formula(lhs("y"),
                        [i(),f(["x0"],nk=10)],
                        data=sim_fit_dat)

    # and the standard deviation as well: log(\sigma_i) = \alpha + f(x0)
    sim1_formula_sd = Formula(lhs("y"),
                        [i(),f(["x0"],nk=10,by="cond")],
                        data=sim_fit_dat)

    # Collect both formulas
    sim1_formulas = [sim1_formula_m,sim1_formula_sd]

    sim_fit_model = GAMMLSS(sim1_formulas,family)
    sim_fit_model.fit(extension_method_lam="nesterov2")

    pred_dat1 = pd.DataFrame({"x0":np.linspace(0,1,50),
                          "cond":["a" for _ in range(50)]})

    pred_dat2 = pd.DataFrame({"x0":np.linspace(0,1,50),
                          "cond":["b" for _ in range(50)]})
    
    diff,ci = sim_fit_model.predict_diff(pred_dat2,pred_dat1,1,[1],whole_interval=True,seed=20)

    def test_print_smooth_p(self):
        capture = io.StringIO()
        with redirect_stdout(capture):
            self.sim_fit_model.print_smooth_terms(p_values=True)
        capture = capture.getvalue()

        comp = "\nDistribution parameter: 1\n\nf(['x0']); edf: 7.079 chi^2: 457.328 P(Chi^2 > chi^2) = 0.000e+00 ***\n\nNote: p < 0.001: ***, p < 0.01: **, p < 0.05: *, p < 0.1: . p-values are approximate!\n\nDistribution parameter: 2\n\nf(['x0'],by=cond): a; edf: 1.0 chi^2: 0.056 P(Chi^2 > chi^2) = 0.81318\nf(['x0'],by=cond): b; edf: 2.789 chi^2: 14.381 P(Chi^2 > chi^2) = 0.00194 **\n\nNote: p < 0.001: ***, p < 0.01: **, p < 0.05: *, p < 0.1: . p-values are approximate!\n"
        assert comp == capture

    def test_diff(self):
        assert np.allclose(self.diff,np.array([-0.30135358, -0.26600859, -0.23106805, -0.19676569, -0.16333525,
                                                -0.13101046, -0.10002506, -0.07060065, -0.0428877 , -0.01701087,
                                                0.00690523,  0.02873598,  0.04835675,  0.06565901,  0.08067376,
                                                0.09350389,  0.10425282,  0.11302403,  0.11992095,  0.12504712,
                                                0.12850733,  0.13040725,  0.13085254,  0.12994891,  0.12780203,
                                                0.12451593,  0.12015716,  0.11475508,  0.10833746,  0.10093206,
                                                0.09256666,  0.0832692 ,  0.0730749 ,  0.06202901,  0.05017749,
                                                0.03756631,  0.02424142,  0.01024862, -0.00438762, -0.01968347,
                                                -0.03565969, -0.05233703, -0.06973624, -0.08787804, -0.10675354,
                                                -0.12627384, -0.14633665, -0.16683967, -0.1876806 , -0.20875715]))
    
    def test_ci(self):
        assert np.allclose(self.ci,np.array([0.34629246, 0.30914794, 0.27596426, 0.24705564, 0.22265442,
                                            0.20269553, 0.18668468, 0.1738519 , 0.16380075, 0.15639354,
                                            0.15136071, 0.14811353, 0.14577211, 0.14347928, 0.14117464,
                                            0.13925221, 0.13794476, 0.1371093 , 0.13623457, 0.13471181,
                                            0.13267291, 0.13079453, 0.12955546, 0.12898548, 0.12865308,
                                            0.12792413, 0.12682113, 0.12604402, 0.1261187 , 0.12710645,
                                            0.12858775, 0.12991388, 0.13096505, 0.13233796, 0.13449826,
                                            0.13749826, 0.14098219, 0.14442194, 0.14773426, 0.1515958 ,
                                            0.15668965, 0.16343419, 0.17198806, 0.1824472 , 0.19525605,
                                            0.21136223, 0.23150348, 0.25598021, 0.28474244, 0.31760382]))


from mssm.models import *
import numpy as np
import os
from mssmViz.sim import*
from numdifftools import Gradient,Hessian
from mssm.src.python.gamm_solvers import deriv_transform_mu_eta,deriv_transform_eta_beta

class GAMLSSGENSMOOTHFamily(GENSMOOTHFamily):
    """Implementation of the ``GENSMOOTHFamily`` class that uses only information about the likelihood to estimate
    a GAMLSS model.

    References:

        - Wood, Pya, & Säfken (2016). Smoothing Parameter and Model Selection for General Smooth Models.
        - Nocedal & Wright (2006). Numerical Optimization. Springer New York.
    """

    def __init__(self, pars: int, links:[Link], llkfun:Callable, *llkargs) -> None:
        super().__init__(pars, links, *llkargs)
        self.llkfun = llkfun
    
    def llk(self, coef, coef_split_idx, y, Xs):
        return self.llkfun(coef, coef_split_idx, self.links, y, Xs,*self.llkargs)
    
    def gradient(self, coef, coef_split_idx, y, Xs):
        """
        Function to evaluate gradient for gsmm model.
        """
        coef = coef.reshape(-1,1)
        split_coef = np.split(coef,coef_split_idx)
        eta_mu = Xs[0]@split_coef[0]
        if len(Xs) > 1:
            eta_sd = Xs[1]@split_coef[1]
        
        # Get the Gamlss family
        gammlss_family = self.llkargs[0]
        
        if len(Xs) > 1:
            d1eta,d2eta,d2meta = deriv_transform_mu_eta(y,[self.links[0].fi(eta_mu),self.links[1].fi(eta_sd)],gammlss_family)
        else:
            d1eta,d2eta,d2meta = deriv_transform_mu_eta(y,[self.links[0].fi(eta_mu)],gammlss_family)
            
        grad,_ = deriv_transform_eta_beta(d1eta,d2eta,d2meta,Xs,only_grad=True)
        #print(pgrad.flatten())
        return grad.reshape(-1,1)
    
    def hessian(self, coef, coef_split_idx, y, Xs):
        return None
            

def llk_gamm_fun(coef,coef_split_idx,links,y,Xs,gammlss_family):
    """Likelihood for a GAM(LSS) - implemented so
    that the model can be estimated using the general smooth code.

    Note, gammlss_family is passed via llkargs so that this code works with
    Gaussian and Gamma models.
    """
    coef = coef.reshape(-1,1)
    split_coef = np.split(coef,coef_split_idx)
    eta_mu = Xs[0]@split_coef[0]
    if len(Xs) > 1:
        eta_sd = Xs[1]@split_coef[1]
    
    mu_mu = links[0].fi(eta_mu)
    if len(Xs) > 1:
        mu_sd = links[1].fi(eta_sd)
    
    if len(Xs) > 1:
        llk = gammlss_family.llk(y,mu_mu,mu_sd)
    else:
        llk = gammlss_family.llk(y,mu_mu)

    if np.isnan(llk):
        return -np.inf
    
    return llk

################################################################## Tests ##################################################################

class Test_GAUMLSSGEN:

    # Simulate 500 data points
    sim_dat = sim3(500,2,c=1,seed=0,family=Gaussian(),binom_offset = 0, correlate=False)

    # We need to model the mean: \mu_i
    formula_m = Formula(lhs("y"),
                        [i(),f(["x0"]),f(["x1"]),f(["x2"]),f(["x3"])],
                        data=sim_dat)

    # And for sd - here constant
    formula_sd = Formula(lhs("y"),
                        [i()],
                        data=sim_dat)

    # Collect both formulas
    formulas = [formula_m,formula_sd]
    links = [Identity(),LOG()]

    # Now define the general family + model and fit!
    gsmm_fam = GAMLSSGENSMOOTHFamily(2,links,llk_gamm_fun,GAUMLSS(links))
    model = GSMM(formulas=formulas,family=gsmm_fam)

    # First fit with SR1 and, to speed things up, only then with Newton
    bfgs_opt={"gtol":1e-9,
            "ftol":1e-9,
            "maxcor":30,
            "maxls":200,
            "maxfun":1e7}
                    
    model.fit(init_coef=None,method='qEFS',extend_lambda=False,
            control_lambda=False,max_outer=200,max_inner=500,min_inner=500,
            seed=0,qEFSH='SR1',max_restarts=5,overwrite_coef=False,qEFS_init_converge=False,prefit_grad=True,
            progress_bar=True,**bfgs_opt)

    model2 = copy.deepcopy(model)

    # Now fit with BFGS
    model2.fit(init_coef=None,method='qEFS',extend_lambda=False,
            control_lambda=False,max_outer=200,max_inner=500,min_inner=500,
            seed=0,qEFSH='BFGS',max_restarts=0,overwrite_coef=True,qEFS_init_converge=True,prefit_grad=False,
            progress_bar=True,**bfgs_opt)

    def test_GAMedf(self):
        assert round(self.model.edf,ndigits=3) == 19.049

    def test_GAMedf2(self):
        assert round(self.model2.edf,ndigits=3) == 16.305

    def test_GAMcoef(self):
        coef = self.model.overall_coef.flatten()
        assert np.allclose(coef,np.array([ 7.64905998, -0.86968557, -0.83494213,  0.07122906,  1.20919857,
                                            1.67258359,  0.46597024, -0.53378393, -0.76055135, -0.7941819 ,
                                            -1.81015845, -1.07691283, -0.44119683,  0.33501224,  1.01000229,
                                            1.9723699 ,  3.16101178,  4.12055134,  5.27092295, -9.19693724,
                                            4.33968853,  5.20719195, -2.48498028, -0.41483633, -1.68072058,
                                            -4.12405797, -4.32178097, -1.28126424, -0.05013006, -0.01078947,
                                            0.01309918,  0.0304733 ,  0.04832851,  0.06789398,  0.08949696,
                                            0.09934445,  0.10436846,  0.66643841]))

    def test_GAMcoef2(self):
        coef = self.model2.overall_coef.flatten()
        assert np.allclose(coef,np.array([  7.6489859 ,  -0.90359436,  -1.1995978 ,  -0.1219847 ,
                                            0.8902022 ,   1.53926404,   0.09832465,  -0.914336  ,
                                            -0.87410618,  -0.50078609,  -1.70675555,  -1.40414449,
                                            -0.78427345,   0.28350426,   0.78126954,   1.56534334,
                                            2.89252623,   4.27551005,   5.81263687, -10.35107199,
                                            3.37333916,   4.0633653 ,  -3.53414945,  -1.49195269,
                                            -2.73210049,  -5.26428994,  -4.80803947,  -2.02328384,
                                            -0.01671727,  -0.02500594,  -0.0339186 ,  -0.0572995 ,
                                            -0.03603238,   0.02741854,   0.08237743,   0.14099506,
                                            0.20398593,   0.66054218]))  

    def test_GAMlam(self):
        lam = np.array([p.lam for p in self.model.overall_penalties])
        assert np.allclose(lam,np.array([7.21928261e-01, 5.68402996e+00, 7.31895390e-03, 1.00000000e+07]))

    def test_GAMlam2(self):
        lam = np.array([p.lam for p in self.model2.overall_penalties])
        assert np.allclose(lam,np.array([2.61309258e-01, 9.35511512e-01, 5.32567098e-03, 3.27034646e+01]))  

    def test_GAMreml(self):
        reml = self.model.get_reml()
        assert round(reml,ndigits=3) == -1078.492

    def test_GAMreml2(self):
        reml = self.model2.get_reml()
        assert round(reml,ndigits=3) == -1062.438

    def test_GAMllk(self):
        llk = self.model.get_llk(False)
        assert round(llk,ndigits=3) == -1042.691
    
    def test_GAMllk2(self):
        llk = self.model2.get_llk(False)
        assert round(llk,ndigits=3) == -1039.741


class Test_PropHaz:
    sim_dat = sim3(500,2,c=1,seed=0,family=PropHaz([0],[0]),binom_offset = 0.1,correlate=False)
        
    # Prep everything for prophaz model
    sim_dat = sim_dat.sort_values(['y'],ascending=[False])
    sim_dat = sim_dat.reset_index(drop=True)
    #print(sim_dat.head(),np.mean(sim_dat["delta"]))

    u,inv = np.unique(sim_dat["y"],return_inverse=True)
    ut = np.flip(u)
    r = np.abs(inv - max(inv))

    # We only need to model the mean: \mu_i
    sim_formula_m = Formula(lhs("delta"),
                        [f(["x0"]),f(["x1"]),f(["x2"]),f(["x3"])],
                        data=sim_dat)

    # Fit with Newton
    gsmm_newton_fam = PropHaz(ut,r)
    gsmm_newton = GSMM([copy.deepcopy(sim_formula_m)],gsmm_newton_fam)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        gsmm_newton.fit(init_coef=None,method="QR/Chol",extend_lambda=False,
                        control_lambda=False,max_outer=200,seed=0,max_inner=500,
                        min_inner=500,progress_bar=True)
        

    gsmm_qefs_fam = PropHaz(ut,r)
    gsmm_qefs = GSMM([copy.deepcopy(sim_formula_m)],gsmm_qefs_fam)

    # Fit with qEFS update without initialization
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        bfgs_opt={"gtol":1e-9,
                    "ftol":1e-9,
                    "maxcor":30,
                    "maxls":200,
                    "maxfun":1e7}
        
        gsmm_qefs.fit(init_coef=None,method='qEFS',extend_lambda=False,
                        control_lambda=False,max_outer=200,max_inner=500,min_inner=500,
                        seed=0,qEFSH='SR1',max_restarts=0,overwrite_coef=False,qEFS_init_converge=False,prefit_grad=True,
                        progress_bar=True,**bfgs_opt)

    def test_GAMcoef(self):

        assert np.allclose((self.gsmm_newton.overall_coef - self.gsmm_qefs.overall_coef).flatten(),
                           np.array([ 0.00085094, -0.00112174, -0.00254822, -0.00133955,  0.00105183,
                                        -0.00174351, -0.00296606,  0.00099741,  0.00571504, -0.00296058,
                                        -0.00230932,  0.00126035,  0.00379059,  0.00195697,  0.00087278,
                                        0.00414338, -0.01210031, -0.03304087,  0.03062558,  0.02426205,
                                        0.01574717,  0.03415708,  0.02328558,  0.02059044,  0.02940937,
                                        0.01876927,  0.02136963,  0.00305857, -0.01617596, -0.02171221,
                                        -0.00669945,  0.01213701,  0.0186336 , -0.00328931, -0.04652364,
                                        -0.0910511 ])) 

    def test_GAMlam(self):

        assert np.allclose(np.round(np.array([p1.lam - p2.lam for p1,p2 in zip(self.gsmm_newton.overall_penalties,self.gsmm_qefs.overall_penalties)]),decimals=3),
                           np.array([ 0.282,  0.544,  0.   , 44.97 ])) 

    def test_GAMreml(self):

        assert round(self.gsmm_newton.get_reml() - self.gsmm_qefs.get_reml(),ndigits=3) == np.float64(-0.816)

    def test_GAMllk(self):

        assert round(self.gsmm_newton.get_llk(True) - self.gsmm_qefs.get_llk(True),ndigits=3) == np.float64(-0.498)


class Test_PropHaz_repara:
    sim_dat = sim3(500,2,c=1,seed=0,family=PropHaz([0],[0]),binom_offset = 0.1,correlate=False)
        
    # Prep everything for prophaz model
    sim_dat = sim_dat.sort_values(['y'],ascending=[False])
    sim_dat = sim_dat.reset_index(drop=True)
    #print(sim_dat.head(),np.mean(sim_dat["delta"]))

    u,inv = np.unique(sim_dat["y"],return_inverse=True)
    ut = np.flip(u)
    r = np.abs(inv - max(inv))

    # We only need to model the mean: \mu_i
    sim_formula_m = Formula(lhs("delta"),
                        [f(["x0"]),f(["x1"]),f(["x2"]),f(["x3"])],
                        data=sim_dat)

    # Fit with Newton
    gsmm_newton_fam = PropHaz(ut,r)
    gsmm_newton = GSMM([copy.deepcopy(sim_formula_m)],gsmm_newton_fam)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        gsmm_newton.fit(init_coef=None,method="QR/Chol",extend_lambda=False,
                        control_lambda=False,max_outer=200,seed=0,max_inner=500,
                        min_inner=500,progress_bar=True,repara=True)
        

    gsmm_qefs_fam = PropHaz(ut,r)
    gsmm_qefs = GSMM([copy.deepcopy(sim_formula_m)],gsmm_qefs_fam)

    # Fit with qEFS update without initialization
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        bfgs_opt={"gtol":1e-9,
                    "ftol":1e-9,
                    "maxcor":30,
                    "maxls":200,
                    "maxfun":1e7}
        
        gsmm_qefs.fit(init_coef=None,method='qEFS',extend_lambda=False,
                        control_lambda=False,max_outer=200,max_inner=500,min_inner=500,
                        seed=0,qEFSH='SR1',max_restarts=0,overwrite_coef=False,qEFS_init_converge=False,prefit_grad=True,
                        progress_bar=True,repara=True,**bfgs_opt)

    def test_GAMcoef(self):

        assert np.allclose((self.gsmm_newton.overall_coef - self.gsmm_qefs.overall_coef).flatten(),
                           np.array([ 0.00130659, -0.00146633, -0.00482344, -0.00219018,  0.00408372,
                                    -0.00254493, -0.00538949,  0.00269739,  0.01220294, -0.00574834,
                                    -0.00341937,  0.00293582,  0.00723694,  0.0041163 ,  0.00185322,
                                    0.00810604, -0.0177597 , -0.05120267, -0.0156164 , -0.01146629,
                                    -0.03849618, -0.00090847, -0.02320778, -0.0266449 , -0.01706707,
                                    0.00320521, -0.00176434,  0.00787907, -0.03024446, -0.04263593,
                                    -0.01658554,  0.01727422,  0.0315337 , -0.0050121 , -0.08136764,
                                    -0.15992784])) 

    def test_GAMlam(self):

        assert np.allclose(np.round(np.array([p1.lam - p2.lam for p1,p2 in zip(self.gsmm_newton.overall_penalties,self.gsmm_qefs.overall_penalties)]),decimals=3),
                           np.array([ 0.671,  0.853, -0.   , 63.778])) 

    def test_GAMreml(self):

        assert round(self.gsmm_newton.get_reml() - self.gsmm_qefs.get_reml(),ndigits=3) == -0.865

    def test_GAMllk(self):

        assert round(self.gsmm_newton.get_llk(True) - self.gsmm_qefs.get_llk(True),ndigits=3) == -0.758


from mssm.models import *
from mssm.src.python.compare import compare_CDL
from mssm.src.python.gamm_solvers import compute_S_emb_pinv_det,cpp_dChol,cpp_chol
from mssm.src.python.utils import estimateVp,correct_VB,DummyRhoPrior
import numpy as np
import os
from mssmViz.sim import*

class Test_model_comparisons1_hard:

    # Model comparison and smoothness uncertainty correction tests

    # Simulate some data
    sim_fit_dat = sim3(n=500,scale=2,c=0.1,family=Gaussian(),seed=21)

    # Now fit nested models
    sim_fit_formula = Formula(lhs("y"),
                                [i(),f(["x0"],nk=20,rp=1),f(["x1"],nk=20,rp=1),f(["x2"],nk=20,rp=1),f(["x3"],nk=20,rp=1)],
                                data=sim_fit_dat,
                                print_warn=False)

    sim_fit_model = GAMM(sim_fit_formula,Gaussian())
    sim_fit_model.fit(exclude_lambda=False,progress_bar=False,max_outer=100)

    sim_fit_formula2 = Formula(lhs("y"),
                                [i(),f(["x1"],nk=20,rp=1),f(["x2"],nk=20,rp=1),f(["x3"],nk=20,rp=1)],
                                data=sim_fit_dat,
                                print_warn=False)

    sim_fit_model2 = GAMM(sim_fit_formula2,Gaussian())
    sim_fit_model2.fit(exclude_lambda=False,progress_bar=False,max_outer=100)

    prior = DummyRhoPrior(b=np.log(1e12))

    # And perform a couple of bias corrected / smoothness uncertainty corrected / or not / comparisons...
    uncor_result = compare_CDL(sim_fit_model,sim_fit_model2,correct_V=False,correct_t1=False,seed=22)
    cor_result1 = compare_CDL(sim_fit_model,sim_fit_model2,correct_t1=False,n_c=1,grid='JJJ1',seed=22,Vp_fidiff=True)
    cor_result2 = compare_CDL(sim_fit_model,sim_fit_model2,correct_t1=False,n_c=1,grid='JJJ2',seed=22,only_expected_edf=True,use_importance_weights=False,Vp_fidiff=False)
    cor_result3 = compare_CDL(sim_fit_model,sim_fit_model2,correct_t1=False,n_c=10,grid='JJJ2',seed=22,use_importance_weights=False,Vp_fidiff=False)
    cor_result4 = compare_CDL(sim_fit_model,sim_fit_model2,correct_t1=False,n_c=1,grid='JJJ3',seed=22,Vp_fidiff=False,use_importance_weights=True)
    cor_result5 = compare_CDL(sim_fit_model,sim_fit_model2,correct_t1=False,n_c=1,grid='JJJ3',seed=22,Vp_fidiff=False,use_importance_weights=True,prior=prior)

    unbiased_uncor_result = compare_CDL(sim_fit_model,sim_fit_model2,correct_V=False,correct_t1=True,n_c=1,grid='JJJ3',seed=22)
    unbiased_cor_result1 = compare_CDL(sim_fit_model,sim_fit_model2,correct_t1=True,n_c=1,grid='JJJ3',seed=22,use_importance_weights=True,prior=prior)

    def test_comp1(self):
        assert round(self.uncor_result['aic_diff'],ndigits=3) == 1.945

    def test_comp2(self):
        assert round(self.cor_result1['aic_diff'],ndigits=3) == 1.946

    def test_comp3(self):
        assert round(self.cor_result2['aic_diff'],ndigits=3) == 1.668

    def test_comp4(self):
        assert round(self.cor_result3['aic_diff'],ndigits=3) == 1.701

    def test_comp5(self):
        assert round(self.cor_result4['aic_diff'],ndigits=3) == 1.946

    def test_comp6(self):
        assert round(self.cor_result5['aic_diff'],ndigits=3) == 2.942

    def test_comp7(self):
        assert round(self.unbiased_uncor_result['aic_diff'],ndigits=3) == 1.945

    def test_comp8(self):
        assert round(self.unbiased_cor_result1['aic_diff'],ndigits=3) == 2.942
    
    def test_edf1(self):
        assert round(self.unbiased_cor_result1['Res. DOF'],ndigits=3) == 0.981


class Test_model_comparisons2_hard:

    # Model comparison and smoothness uncertainty correction tests

    # Simulate some data
    sim_fit_dat = sim3(n=500,scale=2,c=0.1,family=Gamma(),seed=21)

    # Now fit nested models
    sim_fit_formula = Formula(lhs("y"),
                                [i(),f(["x0"],nk=20,rp=1),f(["x1"],nk=20,rp=1),f(["x2"],nk=20,rp=1),f(["x3"],nk=20,rp=1)],
                                data=sim_fit_dat,
                                print_warn=False)

    sim_fit_model = GAMM(sim_fit_formula,Gamma())
    sim_fit_model.fit(exclude_lambda=False,progress_bar=False,max_outer=100)

    sim_fit_formula2 = Formula(lhs("y"),
                                [i(),f(["x1"],nk=20,rp=1),f(["x2"],nk=20,rp=1),f(["x3"],nk=20,rp=1)],
                                data=sim_fit_dat,
                                print_warn=False)

    sim_fit_model2 = GAMM(sim_fit_formula2,Gamma())
    sim_fit_model2.fit(exclude_lambda=False,progress_bar=False,max_outer=100)

    prior = DummyRhoPrior(b=np.log(1e12))
    # And perform a couple of bias corrected / smoothness uncertainty corrected / or not / comparisons...
    uncor_result = compare_CDL(sim_fit_model,sim_fit_model2,correct_V=False,correct_t1=False,seed=22)
    cor_result1 = compare_CDL(sim_fit_model,sim_fit_model2,correct_t1=False,n_c=1,grid='JJJ1',seed=22,Vp_fidiff=True)
    cor_result2 = compare_CDL(sim_fit_model,sim_fit_model2,correct_t1=False,n_c=1,grid='JJJ2',seed=22,only_expected_edf=True,use_importance_weights=False,Vp_fidiff=False)
    cor_result3 = compare_CDL(sim_fit_model,sim_fit_model2,correct_t1=False,n_c=10,grid='JJJ2',seed=22,use_importance_weights=False,Vp_fidiff=False)
    cor_result4 = compare_CDL(sim_fit_model,sim_fit_model2,correct_t1=False,n_c=1,grid='JJJ3',seed=22,Vp_fidiff=False,use_importance_weights=True)
    cor_result5 = compare_CDL(sim_fit_model,sim_fit_model2,correct_t1=False,n_c=10,grid='JJJ3',seed=22,Vp_fidiff=False,use_importance_weights=True,prior=prior)

    unbiased_uncor_result = compare_CDL(sim_fit_model,sim_fit_model2,correct_V=False,correct_t1=True,n_c=10,grid='JJJ3',seed=22)
    unbiased_cor_result1 = compare_CDL(sim_fit_model,sim_fit_model2,correct_t1=True,n_c=10,grid='JJJ3',seed=22,use_importance_weights=True,prior=prior)

    def test_comp1(self):
        assert round(self.uncor_result['aic_diff'],ndigits=3) == 25.323

    def test_comp2(self):
        assert round(self.cor_result1['aic_diff'],ndigits=3) == 25.149

    def test_comp3(self):
        assert round(self.cor_result2['aic_diff'],ndigits=3) == 24.838

    def test_comp4(self):
        assert round(self.cor_result3['aic_diff'],ndigits=3) == 24.819

    def test_comp5(self):
        assert round(self.cor_result4['aic_diff'],ndigits=3) == 25.058

    def test_comp6(self):
        assert round(self.cor_result5['aic_diff'],ndigits=3) == 25.058

    def test_comp7(self):
        assert round(self.unbiased_uncor_result['aic_diff'],ndigits=3) == 25.323

    def test_comp8(self):
        assert round(self.unbiased_cor_result1['aic_diff'],ndigits=3) == 25.058
    
    def test_edf1(self):
        assert round(self.unbiased_cor_result1['Res. DOF'],ndigits=3) == 22.091

class Test_model_comparison3_hard:

    # Model comparison and smoothness uncertainty correction tests

    # Simulate some data
    sim_fit_dat = sim3(n=500,scale=2,c=0.1,family=Gamma(),seed=21)

    # Now fit nested models
    sim_fit_formula = Formula(lhs("y"),
                                [i(),f(["x0"],nk=20,rp=1),f(["x1"],nk=20,rp=1),f(["x2"],nk=20,rp=1),f(["x3"],nk=20,rp=1)],
                                data=sim_fit_dat,
                                print_warn=False)

    sim_fit_formula_sd = Formula(lhs("y"),
                                [i()],
                                data=sim_fit_dat,
                                print_warn=False)

    sim_fit_model = GAMMLSS([sim_fit_formula,copy.deepcopy(sim_fit_formula_sd)],family = GAMMALS([LOG(),LOGb(-0.01)]))
    sim_fit_model.fit(progress_bar=False,max_outer=200,extend_lambda=False,control_lambda=2,max_inner=500)

    sim_fit_formula2 = Formula(lhs("y"),
                                [i(),f(["x1"],nk=20,rp=1),f(["x2"],nk=20,rp=1),f(["x3"],nk=20,rp=1)],
                                data=sim_fit_dat,
                                print_warn=False)

    sim_fit_model2 = GAMMLSS([sim_fit_formula2,copy.deepcopy(sim_fit_formula_sd)],family = GAMMALS([LOG(),LOGb(-0.01)]))
    sim_fit_model2.fit(progress_bar=False,max_outer=200,extend_lambda=False,control_lambda=2,max_inner=500)

    prior = DummyRhoPrior(b=np.log(1e12))
    # And perform a couple of bias corrected / smoothness uncertainty corrected / or not / comparisons...
    uncor_result = compare_CDL(sim_fit_model,sim_fit_model2,correct_V=False,correct_t1=False,seed=22)
    cor_result1 = compare_CDL(sim_fit_model,sim_fit_model2,correct_t1=False,n_c=1,grid='JJJ1',seed=22,Vp_fidiff=True)
    cor_result2 = compare_CDL(sim_fit_model,sim_fit_model2,correct_t1=False,n_c=1,grid='JJJ2',seed=22,only_expected_edf=True,use_importance_weights=False,Vp_fidiff=False)
    cor_result3 = compare_CDL(sim_fit_model,sim_fit_model2,correct_t1=False,n_c=1,grid='JJJ2',seed=22,use_importance_weights=False,Vp_fidiff=False)
    cor_result4 = compare_CDL(sim_fit_model,sim_fit_model2,correct_t1=False,n_c=1,grid='JJJ3',seed=22,Vp_fidiff=False,use_importance_weights=True)
    cor_result5 = compare_CDL(sim_fit_model,sim_fit_model2,correct_t1=False,n_c=1,grid='JJJ3',seed=22,Vp_fidiff=False,use_importance_weights=True,prior=prior)

    unbiased_uncor_result = compare_CDL(sim_fit_model,sim_fit_model2,correct_V=False,correct_t1=True,n_c=1,grid='JJJ3',seed=22)
    unbiased_cor_result1 = compare_CDL(sim_fit_model,sim_fit_model2,correct_t1=True,n_c=1,grid='JJJ3',seed=22,use_importance_weights=True,prior=prior)

    def test_comp1(self):
        assert round(self.uncor_result['aic_diff'],ndigits=3) == 1.912

    def test_comp2(self):
        assert round(self.cor_result1['aic_diff'],ndigits=3) == 2.799

    def test_comp3(self):
        assert round(self.cor_result2['aic_diff'],ndigits=3) == 2.212

    def test_comp4(self):
        assert round(self.cor_result3['aic_diff'],ndigits=3) == 2.195

    def test_comp5(self):
        assert round(self.cor_result4['aic_diff'],ndigits=3) == 1.945

    def test_comp6(self):
        assert round(self.cor_result5['aic_diff'],ndigits=3) == 2.832

    def test_comp7(self):
        assert round(self.unbiased_uncor_result['aic_diff'],ndigits=3) == 1.912

    def test_comp8(self):
        assert round(self.unbiased_cor_result1['aic_diff'],ndigits=3) == 2.832
    
    def test_edf1(self):
        assert round(self.unbiased_cor_result1['Res. DOF'],ndigits=3) == 0.974


class Test_Vb_corrections:
    # Simulate some data - effect of x0 is very very small
    sim_fit_dat = sim3(n=500,scale=2,c=1,family=Gaussian(),seed=21)

    # Now fit nested models
    sim_fit_formula = Formula(lhs("y"),
                                [i(),f(["x0"],nk=20,rp=0),f(["x1"],nk=20,rp=0),f(["x2"],nk=20,rp=0),f(["x3"],nk=20,rp=0)],
                                data=sim_fit_dat,
                                print_warn=False)

    model = GAMM(sim_fit_formula,Gaussian())
    model.fit(exclude_lambda=False,progress_bar=False,max_outer=100)

    # Now fit nested models
    sim_fit_formula2 = Formula(lhs("y"),
                                [i(),f(["x0"],nk=10,rp=0),f(["x1"],nk=10,rp=0),f(["x2"],nk=10,rp=0)],
                                data=sim_fit_dat,
                                print_warn=False)

    model2 = GAMM(sim_fit_formula2,Gaussian())
    model2.fit(exclude_lambda=False,progress_bar=False,max_outer=100)

    Vp1,_,_,_,_ = estimateVp(model,grid_type="JJJ1",verbose=True,seed=20)

    _,_,Vp2,_,_,total_edf,_,_,_,_ = correct_VB(model,grid_type="JJJ1",verbose=True,seed=20,df=2)

    _,LI,_,_,_,total_edf2,_,_,_,_ = correct_VB(model2,grid_type="JJJ2",verbose=True,seed=20,df=2)

    def test_Vp1(self):
        assert np.allclose(np.round(self.Vp1,decimals=3),np.array([[ 1.134000e+00, -7.000000e-03,  7.000000e-03, -1.050000e-01],
                                                                    [-7.000000e-03,  8.970000e-01,  1.600000e-02, -7.000000e-03],
                                                                    [ 7.000000e-03,  1.600000e-02,  4.470000e-01,  9.600000e-02],
                                                                    [-1.050000e-01, -7.000000e-03,  9.600000e-02,  5.989569e+03]]))
    
    def test_Vp2(self):
        assert np.allclose(self.Vp1,self.Vp2)
    
    def test_edf1(self):
        assert np.round(self.total_edf,decimals=3) == 22.15

    def test_edf2(self):
        assert np.round(self.total_edf2,decimals=3) == 16.857


from setuptools import setup
import git
from pathlib import Path
from pybind11.setup_helpers import Pybind11Extension
import os

# Eigen can either be provided by conda or needs to be in "usr/include/eigen3"
# ToDo: The latter might break under windows.

class get_eigen_include(object):
    # Based on the implementation from: https://github.com/MatPiq/RlassoModels/blob/master/setup.py
    # This finally, allowed me to fix my building issues since the author of the repo above had
    # the same problem: https://github.com/readthedocs/readthedocs.org/issues/9034.
    # I made some changes to stream-line the cloning into Eigen a bit.

    def __str__(self) -> str:

        # Local build based on conda
        if "CONDA_PREFIX" in os.environ:
            if os.path.isdir(os.environ["CONDA_PREFIX"] + "/include/eigen3"):
                return os.environ["CONDA_PREFIX"] + "/include/eigen3"

        # Local build with alternative accepted location.
        if os.path.isdir("usr/local/include/eigen3"):
            return "usr/local/include/eigen3"

        target_dir = Path(__file__).resolve().parent / "eigen"
        if target_dir.exists():
            return target_dir.name

        # If we cannot find an eigen installation in the current step we have to download it first.
        # See: https://gist.github.com/plembo/a786ce2851cec61ac3a051fcaf3ccdab
        git.Repo.clone_from("https://gitlab.com/libeigen/eigen.git",target_dir,branch="3.4.0")

        return target_dir.name

# Create Pybind setuptools extension
ext = Pybind11Extension(name='cpp_solvers',
                        sources=['src/mssm/src/cpp/cpp_solvers.cpp'],
                        include_dirs=[get_eigen_include()],
                        cxx_std=14)

setup(ext_modules=[ext])
